"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import logging
import random
import re
import string
from typing import TYPE_CHECKING, Any, TypeVar

import aiohttp
import discord
from discord.enums import try_enum
from discord.utils import MISSING, classproperty
import urllib.parse

from .enums import LoadType, NodeStatus
from .exceptions import *
from .websocket import Websocket

if TYPE_CHECKING:
    from .player import Player
    from .tracks import *
    from .types.request import Request
    from .ext import spotify as spotify_

    PlayableT = TypeVar('PlayableT', bound=Playable)
    

__all__ = ('Node', 'NodePool')


logger: logging.Logger = logging.getLogger(__name__)


# noinspection PyShadowingBuiltins
class Node:
    """The base Wavelink Node.

    The Node is responsible for keeping the Websocket alive, tracking the state of Players
    and fetching/decoding Tracks and Playlists.

    .. note::

        The Node class should only be created once per Lavalink connection.
        To retrieve a Node use the appropriate :class:`NodePool` methods instead.

    .. warning::

        The Node will not be connected until passed to :meth:`NodePool.connect`.

    Parameters
    ----------
    id: Optional[str]
        The unique identifier for this Node. If not passed, one will be generated randomly.
    uri: str
        The uri to connect to your Lavalink server. E.g ``http://localhost:2333``.
    password: str
        The password used to connect to your Lavalink server.
    secure: Optional[bool]
        Whether the connection should use https/wss.
    use_http: Optional[bool]
        Whether to use http:// over ws:// when connecting to the Lavalink websocket. Defaults to False.
    session: Optional[aiohttp.ClientSession]
        The session to use for this Node. If no session is passed a default will be used.
    heartbeat: float
        The time in seconds to send a heartbeat ack. Defaults to 15.0.
    retries: Optional[int]
        The amount of times this Node will try to reconnect after a disconnect.
        If not set the Node will try unlimited times.

    Attributes
    ----------
    heartbeat: float
        The time in seconds to send a heartbeat ack. Defaults to 15.0.
    client: :class:`discord.Client`
        The discord client used to connect this Node. Could be None if this Node has not been connected.
    """

    def __init__(
            self,
            *,
            id: str | None = None,
            uri: str,
            password: str,
            secure: bool = False,
            use_http: bool = False,
            session: aiohttp.ClientSession = MISSING,
            heartbeat: float = 15.0,
            retries: int | None = None,
    ) -> None:
        if id is None:
            id = ''.join(random.sample(string.ascii_letters + string.digits, 12))

        self._id: str = id
        self._uri: str = uri
        self._secure: bool = secure
        self._use_http: bool = use_http
        host: str = re.sub(r'(?:http|ws)s?://', '', self._uri)
        self._host: str = f'{"https://" if secure else "http://"}{host}'
        self._password: str = password

        self._session: aiohttp.ClientSession = session
        self.heartbeat: float = heartbeat
        self._retries: int | None = retries

        self.client: discord.Client | None = None
        self._websocket: Websocket = MISSING
        self._session_id: str | None = None

        self._players: dict[int, Player] = {}

        self._status: NodeStatus = NodeStatus.DISCONNECTED
        self._major_version: int | None = None

        self._spotify: spotify_.SpotifyClient | None = None

    def __repr__(self) -> str:
        return f'Node: id="{self._id}", uri="{self.uri}", status={self.status}'

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, Node) else NotImplemented

    @property
    def id(self) -> str:
        """The Nodes unique identifier."""
        return self._id

    @property
    def uri(self) -> str:
        """The URI used to connect this Node to Lavalink."""
        return self._host

    @property
    def password(self) -> str:
        """The password used to connect this Node to Lavalink."""
        return self._password

    @property
    def players(self) -> dict[int, Player]:
        """A mapping of Guild ID to Player."""
        return self._players

    @property
    def status(self) -> NodeStatus:
        """The connection status of this Node.

        DISCONNECTED
        CONNECTING
        CONNECTED
        """
        return self._status

    def get_player(self, guild_id: int, /) -> Player | None:
        """Return the :class:`player.Player` associated with the provided guild ID.

        If no :class:`player.Player` is found, returns None.

        Parameters
        ----------
        guild_id: int
            The Guild ID to return a Player for.

        Returns
        -------
        Optional[:class:`player.Player`]
        """
        return self._players.get(guild_id, None)

    async def _connect(self, client: discord.Client) -> None:
        if client.user is None:
            raise RuntimeError('')

        if not self._session:
            self._session = aiohttp.ClientSession(headers={'Authorization': self._password})

        self.client = client

        self._websocket = Websocket(node=self)

        await self._websocket.connect()

        async with self._session.get(f'{self._host}/version') as resp:
            version: str = await resp.text()

            if version.endswith('-SNAPSHOT'):
                self._major_version = 3
                return

            version_tuple = tuple(int(v) for v in version.split('.'))
            if version_tuple[0] < 3:
                raise InvalidLavalinkVersion(f'Wavelink 2 is not compatible with Lavalink "{version}".')

            if version_tuple[0] == 3 and version_tuple[1] < 7:
                raise InvalidLavalinkVersion('Wavelink 2 is not compatible with Lavalink versions under "3.7".')

            self._major_version = version_tuple[0]

    async def _send(self,
                    *,
                    method: str,
                    path: str,
                    guild_id: int | str | None = None,
                    query: str | None = None,
                    data: Request | None = None,
                    ) -> dict[str, Any] | None:

        uri: str = f'{self._host}/' \
                   f'v{self._major_version}/' \
                   f'{path}' \
                   f'{f"/{guild_id}" if guild_id else ""}' \
                   f'{f"?{query}" if query else ""}'

        async with self._session.request(method=method, url=uri, json=data or {}) as resp:
            if resp.status >= 300:
                raise InvalidLavalinkResponse(f'An error occurred when attempting to reach: "{uri}".',
                                              status=resp.status)

            if resp.status == 204:
                return

            return await resp.json()

    async def get_tracks(self, cls: type[PlayableT], query: str) -> list[PlayableT]:
        """|coro|

        Search for and retrieve Tracks based on the query and cls provided.

        .. note::

            If the query is not a Local search or direct URL, you will need to provide a search prefix.
            E.g. ``ytsearch:`` for a YouTube search.

        Parameters
        ----------
        cls: type[PlayableT]
            The type of Playable tracks that should be returned.
        query: str
            The query to search for and return tracks.

        Returns
        -------
        list[PlayableT]
            A list of found tracks converted to the provided cls.
        """
        data = await self._send(method='GET', path='loadtracks', query=f'identifier={query}')
        load_type = try_enum(LoadType, data.get("loadType"))

        if load_type is LoadType.load_failed:
            # TODO - Proper Exception...

            raise ValueError('Track Failed to load.')

        if load_type is LoadType.no_matches:
            return []

        if load_type is LoadType.track_loaded:
            track_data = data["tracks"][0]
            return [cls(track_data)]

        if load_type is not LoadType.search_result:
            # TODO - Proper Exception...

            raise ValueError('Track Failed to load.')

        return [cls(track_data) for track_data in data["tracks"]]

    async def get_playlist(self, cls: Playlist, query: str):
        """|coro|

        Search for and return a :class:`tracks.Playlist` given an identifier.

        Parameters
        ----------
        cls: Type[:class:`tracks.Playlist`]
            The type of which playlist should be returned, this must subclass :class:`tracks.Playlist`.
        query: str
            The playlist's identifier. This may be a YouTube playlist URL for example.

        Returns
        -------
        Optional[:class:`tracks.Playlist`]:
            The related wavelink track object or ``None`` if none was found.

        Raises
        ------
        ValueError
            Loading the playlist failed.
        WavelinkException
            An unspecified error occurred when loading the playlist.
        """
        encoded_query = urllib.parse.quote(query)
        data = await self._send(method='GET', path='loadtracks', query=f'identifier={encoded_query}')

        load_type = try_enum(LoadType, data.get("loadType"))

        if load_type is LoadType.load_failed:
            # TODO Proper exception...
            raise ValueError('Tracks failed to Load.')

        if load_type is LoadType.no_matches:
            return None

        if load_type is not LoadType.playlist_loaded:
            raise WavelinkException("Track failed to load.")

        return cls(data)

    async def build_track(self, *, cls: type[PlayableT], encoded: str) -> PlayableT:
        """|coro|

        Build a track from the provided encoded string with the given Track class.

        Parameters
        ----------
        cls: type[PlayableT]
            The type of Playable track that should be returned.
        encoded: str
            The Tracks unique encoded string.
        """
        encoded_query = urllib.parse.quote(encoded)
        data = await self._send(method='GET', path='decodetrack', query=f'encodedTrack={encoded_query}')

        return cls(data=data)


# noinspection PyShadowingBuiltins
class NodePool:
    """The Wavelink NodePool is responsible for keeping track of all :class:`Node`.

    Attributes
    ----------
    nodes: dict[str, :class:`Node`]
        A mapping of :class:`Node` identifier to :class:`Node`.


    .. warning::

        This class should never be initialised. All methods are class methods.
    """

    __nodes: dict[str, Node] = {}

    @classmethod
    async def connect(
            cls,
            *,
            client: discord.Client,
            nodes: list[Node],
            spotify: spotify_.SpotifyClient | None = None
    ) -> dict[str, Node]:
        """|coro|

        Connect a list of Nodes.

        Parameters
        ----------
        client: :class:`discord.Client`
            The discord Client or Bot used to connect the Nodes.
        nodes: list[:class:`Node`]
            A list of Nodes to connect.
        spotify: Optional[:class:`ext.spotify.SpotifyClient`]
            The spotify Client to use when searching for Spotify Tracks.

        Returns
        -------
        dict[str, :class:`Node`]
            A mapping of :class:`Node` identifier to :class:`Node`.
        """
        if client.user is None:
            raise RuntimeError('')

        for node in nodes:

            if spotify:
                node._spotify = spotify

            if node.id in cls.__nodes:
                logger.error(f'A Node with the ID "{node.id}" already exists on the NodePool. Disregarding.')
                continue

            try:
                await node._connect(client)
            except AuthorizationFailed:
                logger.error(f'The Node <{node!r}> failed to authenticate properly. '
                             f'Please check your password and try again.')
            else:
                cls.__nodes[node.id] = node

        return cls.nodes

    @classproperty
    def nodes(cls) -> dict[str, Node]:
        """A mapping of :class:`Node` identifier to :class:`Node`."""
        return cls.__nodes

    @classmethod
    def get_node(cls, id: str | None = None) -> Node:
        """Retrieve a :class:`Node` with the given ID or best, if no ID was passed.

        Parameters
        ----------
        id: Optional[str]
            The unique identifier of the :class:`Node` to retrieve. If not passed the best :class:`Node`
            will be fetched.

        Returns
        -------
        :class:`Node`

        Raises
        ------
        InvalidNode
            The given id does nto resolve to a :class:`Node` or no :class:`Node` has been connected.
        """
        if id:
            if id not in cls.__nodes:
                raise InvalidNode(f'A Node with ID "{id}" does not exist on the Wavelink NodePool.')

            return cls.__nodes[id]

        if not cls.__nodes:
            raise InvalidNode('No Node currently exists on the Wavelink NodePool.')

        nodes = cls.__nodes.values()
        return sorted(nodes, key=lambda n: len(n.players))[0]

    @classmethod
    def get_connected_node(cls) -> Node:
        """Get the best available connected :class:`Node`.

        Returns
        -------
        :class:`Node`
            The best available connected Node.

        Raises
        ------
        InvalidNode
            No Nodes are currently in the connected state.
        """

        nodes: list[Node] = [n for n in cls.__nodes.values() if n.status is NodeStatus.CONNECTED]
        if not nodes:
            raise InvalidNode('There are no Nodes on the Wavelink NodePool that are currently in the connected state.')

        return sorted(nodes, key=lambda n: len(n.players))[0]

    @classmethod
    async def get_tracks(cls_,  # type: ignore
                         query: str,
                         /,
                         *,
                         cls: type[PlayableT],
                         node: Node | None = None
                         ) -> list[PlayableT]:
        """|coro|

        Helper method to retrieve tracks from the NodePool without fetching a :class:`Node`.

        Parameters
        ----------
        query: str
            The query to search for and return tracks.
        cls: type[PlayableT]
            The type of Playable tracks that should be returned.
        node: Optional[:class:`Node`]
            The node to use for retrieving tracks. If not passed, the best :class:`Node` will be used.
            Defaults to None.

        Returns
        -------
        list[PlayableT]
            A list of found tracks converted to the provided cls.
        """
        if not node:
            node = cls_.get_connected_node()

        return await node.get_tracks(cls=cls, query=query)

    @classmethod
    async def get_playlist(cls_,  # type: ignore
                           query: str,
                           /,
                           *,
                           cls: Playlist,
                           node: Node | None = None
                           ) -> Playlist:
        """|coro|

        Helper method to retrieve a playlist from the NodePool without fetching a :class:`Node`.


        .. warning::

            The only playlist currently supported is :class:`tracks.YouTubePlaylist`.


        Parameters
        ----------
        query: str
            The query to search for and return a playlist.
        cls: type[PlayableT]
            The type of Playlist that should be returned.
        node: Optional[:class:`Node`]
            The node to use for retrieving tracks. If not passed, the best :class:`Node` will be used.
            Defaults to None.

        Returns
        -------
        Playlist
            A Playlist with its tracks.
        """
        if not node:
            node = cls_.get_connected_node()

        return await node.get_playlist(cls=cls, query=query)



from typing import TYPE_CHECKING, Literal, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired, TypeAlias


_TrackStartEventType = Literal["TrackStartEvent"]
_OtherEventOpType = Literal["TrackEndEvent", "TrackExceptionEvent", "TrackStuckEvent", "WebSocketClosedEvent"]

EventType = Literal[_TrackStartEventType, _OtherEventOpType]


class _BaseEventOp(TypedDict):
    op: Literal["event"]
    guildId: str

class TrackStartEvent(_BaseEventOp):
    type: _TrackStartEventType
    encodedTrack: str

class _OtherEventOp(_BaseEventOp):
    type: _OtherEventOpType

EventOp: TypeAlias = "TrackStartEvent | _OtherEventOp"


class PlayerState(TypedDict):
    time: int
    position: NotRequired[int]
    connected: bool
    ping: int


class PlayerUpdateOp(TypedDict):
    guildId: str
    state: PlayerState


"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from discord.enums import Enum

__all__ = ('NodeStatus', 'TrackSource', 'LoadType', 'TrackEventType', 'DiscordVoiceCloseType')


class NodeStatus(Enum):
    """Enum representing the current status of a Node.

    Attributes
    ----------
    DISCONNECTED
        0
    CONNECTING
        1
    CONNECTED
        2
    """

    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2


class TrackSource(Enum):
    """Enum representing the Track Source Type.

    Attributes
    ----------
    YouTube
        0
    YouTubeMusic
        1
    SoundCloud
        2
    Local
        3
    Unknown
        4
    """

    YouTube = 0
    YouTubeMusic = 1
    SoundCloud = 2
    Local = 3
    Unknown = 4


class LoadType(Enum):
    """Enum representing the Tracks Load Type.

    Attributes
    ----------
    track_loaded
        "TRACK_LOADED"
    playlist_loaded
        "PLAYLIST_LOADED"
    search_result
        "SEARCH_RESULT"
    no_matches
        "NO_MATCHES"
    load_failed
        "LOAD_FAILED"
    """
    track_loaded = "TRACK_LOADED"
    playlist_loaded = "PLAYLIST_LOADED"
    search_result = "SEARCH_RESULT"
    no_matches = "NO_MATCHES"
    load_failed = "LOAD_FAILED"


class TrackEventType(Enum):
    """Enum representing the TrackEvent types.

    Attributes
    ----------
    START
        "TrackStartEvent"
    END
        "TrackEndEvent"
    """

    START = 'TrackStartEvent'
    END = 'TrackEndEvent'


class DiscordVoiceCloseType(Enum):
    """Enum representing the various Discord Voice Websocket Close Codes.

    Attributes
    ----------
    CLOSE_NORMAL
        1000
    UNKNOWN_OPCODE
        4001
    FAILED_DECODE_PAYLOAD
        4002
    NOT_AUTHENTICATED
        4003
    AUTHENTICATION_FAILED
        4004
    ALREADY_AUTHENTICATED
        4005
    SESSION_INVALID
        4006
    SESSION_TIMEOUT
        4009
    SERVER_NOT_FOUND
        4011
    UNKNOWN_PROTOCOL
        4012
    DISCONNECTED
        4014
    VOICE_SERVER_CRASHED
        4015
    UNKNOWN_ENCRYPTION_MODE
        4016
    """
    CLOSE_NORMAL = 1000  # Not Discord but standard websocket
    UNKNOWN_OPCODE = 4001
    FAILED_DECODE_PAYLOAD = 4002
    NOT_AUTHENTICATED = 4003
    AUTHENTICATION_FAILED = 4004
    ALREADY_AUTHENTICATED = 4005
    SESSION_INVALID = 4006
    SESSION_TIMEOUT = 4009
    SERVER_NOT_FOUND = 4011
    UNKNOWN_PROTOCOL = 4012
    DISCONNECTED = 4014
    VOICE_SERVER_CRASHED = 4015
    UNKNOWN_ENCRYPTION_MODE = 4016



"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
from collections import deque
from collections.abc import AsyncIterator, Iterable, Iterator
from copy import copy
import random

from .exceptions import QueueEmpty
from .tracks import Playable
from .ext import spotify


__all__ = (
    'BaseQueue',
    'Queue'
)


class BaseQueue:

    def __init__(self) -> None:
        self._queue: deque[Playable, spotify.SpotifyTrack] = deque()

    def __str__(self) -> str:
        """String showing all Playable objects appearing as a list."""
        return str([f"'{t}'" for t in self])

    def __repr__(self) -> str:
        """Official representation with max_size and member count."""
        return (
            f"Wavelink Queue: members={self.count}")

    def __bool__(self) -> bool:
        """Treats the queue as a bool, with it evaluating True when it contains members."""
        return bool(self.count)

    def __call__(self, item: Playable | spotify.SpotifyTrack) -> None:
        """Allows the queue instance to be called directly in order to add a member."""
        self.put(item)

    def __len__(self) -> int:
        """Return the number of members in the queue."""
        return self.count

    def __getitem__(self, index: int) -> Playable | spotify.SpotifyTrack:
        """Returns a member at the given position.

        Does not remove item from queue.
        """
        if not isinstance(index, int):
            raise ValueError("'int' type required.'")

        return self._queue[index]

    def __setitem__(self, index: int, item: Playable | spotify.SpotifyTrack):
        """Inserts an item at given position."""
        if not isinstance(index, int):
            raise ValueError("'int' type required.'")

        self.put_at_index(index, item)

    def __delitem__(self, index: int) -> None:
        """Delete item at given position."""
        self._queue.__delitem__(index)

    def __iter__(self) -> Iterator[Playable | spotify.SpotifyTrack]:
        """Iterate over members in the queue.
        Does not remove items when iterating.
        """
        return self._queue.__iter__()

    def __reversed__(self) -> Iterator[Playable | spotify.SpotifyTrack]:
        """Iterate over members in reverse order."""
        return self._queue.__reversed__()

    def __contains__(self, item: Playable | spotify.SpotifyTrack) -> bool:
        """Check if an item is a member of the queue."""
        return item in self._queue

    def __add__(self, other: Iterable[Playable | spotify.SpotifyTrack]):
        """Return a new queue containing all members.

        The new queue will have the same max_size as the original.
        """
        if not isinstance(other, Iterable):
            raise TypeError(f"Adding with the '{type(other)}' type is not supported.")

        new_queue = self.copy()
        new_queue.extend(other)
        return new_queue

    def __iadd__(self, other: Iterable[Playable] | Playable):
        """Add items to queue."""
        if isinstance(other, (Playable, spotify.SpotifyTrack)):
            self.put(other)

            return self

        if isinstance(other, Iterable):
            self.extend(other)
            return self

        raise TypeError(f"Adding '{type(other)}' type to the queue is not supported.")

    def _get(self) -> Playable | spotify.SpotifyTrack:
        if self.is_empty:
            raise QueueEmpty("No items currently in the queue.")

        return self._queue.popleft()

    def _drop(self) -> Playable | spotify.SpotifyTrack:
        return self._queue.pop()

    def _index(self, item: Playable | spotify.SpotifyTrack) -> int:
        return self._queue.index(item)

    def _put(self, item: Playable | spotify.SpotifyTrack) -> None:
        self._queue.append(item)

    def _insert(self, index: int, item: Playable | spotify.SpotifyTrack) -> None:
        self._queue.insert(index, item)

    @staticmethod
    def _check_playable(item: Playable | spotify.SpotifyTrack) -> Playable | spotify.SpotifyTrack:
        if not isinstance(item, (Playable, spotify.SpotifyTrack)):
            raise TypeError("Only Playable objects are supported.")

        return item

    @classmethod
    def _check_playable_container(cls, iterable: Iterable) -> list[Playable | spotify.SpotifyTrack]:
        iterable = list(iterable)

        for item in iterable:
            cls._check_playable(item)

        return iterable

    @property
    def count(self) -> int:
        """Returns queue member count."""
        return len(self._queue)

    @property
    def is_empty(self) -> bool:
        """Returns True if queue has no members."""
        return not bool(self.count)

    def get(self) -> Playable | spotify.SpotifyTrack:
        """Return next immediately available item in queue if any.

        Raises QueueEmpty if no items in queue.
        """
        if self.is_empty:
            raise QueueEmpty("No items currently in the queue.")

        return self._get()

    def pop(self) -> Playable | spotify.SpotifyTrack:
        """Return item from the right end side of the queue.

        Raises QueueEmpty if no items in queue.
        """
        if self.is_empty:
            raise QueueEmpty("No items currently in the queue.")

        return self._queue.pop()

    def find_position(self, item: Playable | spotify.SpotifyTrack) -> int:
        """Find the position a given item within the queue.
        Raises ValueError if item is not in queue.
        """
        return self._index(self._check_playable(item))

    def put(self, item: Playable | spotify.SpotifyTrack) -> None:
        """Put the given item into the back of the queue."""
        self._put(self._check_playable(item))

    def put_at_index(self, index: int, item: Playable | spotify.SpotifyTrack) -> None:
        """Put the given item into the queue at the specified index."""
        self._insert(index, self._check_playable(item))

    def put_at_front(self, item: Playable | spotify.SpotifyTrack) -> None:
        """Put the given item into the front of the queue."""
        self.put_at_index(0, item)

    def shuffle(self) -> None:
        """Shuffles the queue in place. This does not return anything."""
        random.shuffle(self._queue)

    def extend(self, iterable: Iterable[Playable | spotify.SpotifyTrack], *, atomic: bool = True) -> None:
        """Add the members of the given iterable to the end of the queue.

        If atomic is set to True, no tracks will be added upon any exceptions.
        If atomic is set to False, as many tracks will be added as possible.

        """
        if atomic:
            iterable = self._check_playable_container(iterable)

        for item in iterable:
            self.put(item)

    def copy(self):
        """Create a copy of the current queue including its members."""
        new_queue = self.__class__()
        new_queue._queue = copy(self._queue)

        return new_queue

    def clear(self) -> None:
        """Remove all items from the queue."""
        self._queue.clear()


class Queue(BaseQueue):

    def __init__(self):
        super().__init__()
        self.history: BaseQueue = BaseQueue()

        self._loop: bool = False
        self._loop_all: bool = False

        self._loaded = None
        self._waiters = deque()
        self._finished = asyncio.Event()
        self._finished.set()

    async def __aiter__(self) -> AsyncIterator[Playable | spotify.SpotifyTrack]:
        """Pops members as it iterates the queue, waiting for new members when exhausted.

        Removes items when iterating.
        """
        while True:
            yield await self.get_wait()

    def get(self) -> Playable | spotify.SpotifyTrack:
        return self._get()

    def _get(self) -> Playable | spotify.SpotifyTrack:
        if self.loop and self._loaded:
            return self._loaded

        item = super()._get()
        if self.loop_all and self.is_empty:
            self._queue.extend(self.history._queue)
            self.history.clear()

        self._loaded = item
        self.history.put(item)

        return item

    def _put(self, item: Playable | spotify.SpotifyTrack) -> None:
        super()._put(item)
        self._wakeup_next()

    def _insert(self, index: int, item: Playable | spotify.SpotifyTrack) -> None:
        super()._insert(index, item)
        self._wakeup_next()

    def _wakeup_next(self) -> None:
        while self._waiters:
            waiter = self._waiters.popleft()

            if not waiter.done():
                waiter.set_result(None)
                break

    async def get_wait(self) -> Playable | spotify.SpotifyTrack:
        """|coro|

        Return the next item in queue once available.

        .. note::

            This will wait until an item is available to be retrieved.
        """
        while self.is_empty:
            loop = asyncio.get_event_loop()
            waiter = loop.create_future()

            self._waiters.append(waiter)

            try:
                await waiter
            except:  # noqa
                waiter.cancel()

                try:
                    self._waiters.remove(waiter)
                except ValueError:  # pragma: no branch
                    pass

                if not self.is_empty and not waiter.cancelled():  # pragma: no cover
                    # something went wrong with this waiter, move on to next
                    self._wakeup_next()
                raise

        return self.get()

    async def put_wait(self, item: Playable | spotify.SpotifyTrack) -> None:
        """|coro|

        Put an item into the queue using await.
        """
        self._put(item)
        await asyncio.sleep(0)

    def reset(self) -> None:
        """Clears the state of all queues."""
        self.clear()
        self.history.clear()

        for waiter in self._waiters:
            waiter.cancel()

        self._waiters.clear()

        self._loaded = None
        self._loop = False
        self._loop_all = False

    @property
    def loop(self) -> bool:
        """Whether the queue will loop the currently playing song.

        Can be set to True or False.
        Defaults to False.

        Returns
        -------
        bool
        """
        return self._loop

    @loop.setter
    def loop(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError('The "loop" property can only be set with a bool.')

        self._loop = value

    @property
    def loop_all(self) -> bool:
        """Whether the queue will loop all songs in the history queue.

        Can be set to True or False.
        Defaults to False.

        .. note::

            If `loop` is set to True, this has no effect until `loop` is set to False.

        Returns
        -------
        bool
        """
        return self._loop_all

    @loop_all.setter
    def loop_all(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError('The "loop_all" property can only be set with a bool.')

        self._loop_all = value



from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from discord.types.voice import GuildVoiceState, VoiceServerUpdate
    from typing_extensions import NotRequired, TypeAlias
    
class VoiceState(TypedDict):
    token: str
    endpoint: str
    sessionId: str
    connected: NotRequired[bool]
    ping: NotRequired[int]


class DiscordVoiceState(GuildVoiceState, VoiceServerUpdate):
    ...



"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, ClassVar, Literal, overload, Optional, Any

import aiohttp
import yarl
from discord.ext import commands

from .enums import TrackSource
from .exceptions import NoTracksError
from .node import Node, NodePool

if TYPE_CHECKING:
    from typing_extensions import Self

    from .types.track import Track as TrackPayload

__all__ = (
    'Playable',
    'Playlist',
    'YouTubeTrack',
    'GenericTrack',
    'YouTubeMusicTrack',
    'SoundCloudTrack',
    'YouTubePlaylist'
)


_source_mapping: dict[str, TrackSource] = {
    'youtube': TrackSource.YouTube
}


class Playlist(metaclass=abc.ABCMeta):
    """An ABC that defines the basic structure of a lavalink playlist resource.

    Attributes
    ----------
    data: Dict[str, Any]
        The raw data supplied by Lavalink.
    """

    def __init__(self, data: dict[str, Any]):
        self.data: dict[str, Any] = data


class Playable(metaclass=abc.ABCMeta):
    """Base ABC Track used in all the Wavelink Track types.

    Attributes
    ----------
    data: dict[str, Any]
        The raw data received via Lavalink.
    encoded: str
        The encoded Track string.
    is_seekable: bool
        Whether the Track is seekable.
    is_stream: bool
        Whether the Track is a stream.
    length: int
        The length of the track in milliseconds.
    duration: int
        An alias for length.
    position: int
        The position the track will start in milliseconds. Defaults to 0.
    title: str
        The Track title.
    source: :class:`wavelink.TrackSource`
        The source this Track was fetched from.
    uri: Optional[str]
        The URI of this track. Could be None.
    author: Optional[str]
        The author of this track. Could be None.
    identifier: Optional[str]
        The Youtube/YoutubeMusic identifier for this track. Could be None.
    """

    PREFIX: ClassVar[str] = ''
    
    def __init__(self, data: TrackPayload) -> None:
        self.data: TrackPayload = data
        self.encoded: str = data['encoded']

        info = data['info']
        self.is_seekable: bool = info.get('isSeekable', False)
        self.is_stream: bool = info.get('isStream', False)
        self.length: int = info.get('length', 0)
        self.duration: int = self.length
        self.position: int = info.get('position', 0)

        self.title: str = info.get('title', 'Unknown Title')

        source: str | None = info.get('sourceName')
        self.source: TrackSource = _source_mapping.get(source, TrackSource.Unknown)

        self.uri: str | None = info.get('uri')
        self.author: str | None = info.get('author')
        self.identifier: str | None = info.get('identifier')

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return f'Playable: source={self.source}, title={self.title}'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Playable):
            return self.encoded == other.encoded
        return NotImplemented
    
    @overload
    @classmethod
    async def search(cls,
                     query: str,
                     /,
                     *,
                     node: Node | None = ...
                     ) -> list[Self]:
        ...

    @overload
    @classmethod
    async def search(cls,
                     query: str,
                     /,
                     *,
                     node: Node | None = ...
                     ) -> Self:
        ...

    @overload
    @classmethod
    async def search(cls,
                     query: str,
                     /,
                     *,
                     node: Node | None = ...
                     ) -> Self | list[Self]:
        ...

    @overload
    @classmethod
    async def search(cls,
                     query: str,
                     /,
                     *,
                     node: Node | None = ...
                     ) -> YouTubePlaylist:
        ...

    @classmethod
    async def search(cls,
                     query: str,
                     /,
                     *,
                     node: Node | None = None
                     ) -> Self | list[Self]:
        """Search and retrieve tracks for the given query.

        Parameters
        ----------
        query: str
            The query to search for.
        node: Optional[:class:`wavelink.Node`]
            The node to use when searching for tracks. If no :class:`wavelink.Node` is passed,
            one will be fetched via the :class:`wavelink.NodePool`.
        """

        check = yarl.URL(query)

        if str(check.host) == 'youtube.com' or str(check.host) == 'www.youtube.com' and check.query.get("list") or \
                cls.PREFIX == 'ytpl:':

            playlist = await NodePool.get_playlist(query, cls=YouTubePlaylist, node=node)
            return playlist
        elif check.host:
            tracks = await NodePool.get_tracks(query, cls=cls, node=node)
        else:
            tracks = await NodePool.get_tracks(f'{cls.PREFIX}{query}', cls=cls, node=node)

        return tracks

    @classmethod
    async def convert(cls, ctx: commands.Context, argument: str) -> Self:
        """Converter which searches for and returns the first track.

        Used as a type hint in a
        `discord.py command <https://discordpy.readthedocs.io/en/stable/ext/commands/commands.html>`_.
        """
        results = await cls.search(argument)

        if not results:
            raise commands.BadArgument("Could not find any songs matching that query.")

        if issubclass(cls, YouTubePlaylist):
            return results  # type: ignore

        return results[0]


class GenericTrack(Playable):
    """Generic Wavelink Track.

    Use this track for searching for Local songs or direct URLs.
    """
    ...


class YouTubeTrack(Playable):

    PREFIX: str = 'ytsearch:'

    def __init__(self, data: TrackPayload) -> None:
        super().__init__(data)

        self._thumb: str = f"https://img.youtube.com/vi/{self.identifier}/maxresdefault.jpg"

    @property
    def thumbnail(self) -> str:
        """The URL to the thumbnail of this video.

        .. note::

            Due to YouTube limitations this may not always return a valid thumbnail.
            Use :meth:`.fetch_thumbnail` to fallback.

        Returns
        -------
        str
            The URL to the video thumbnail.
        """
        return self._thumb

    thumb = thumbnail

    async def fetch_thumbnail(self, *, node: Node | None = None) -> str:
        """Fetch the max resolution thumbnail with a fallback if it does not exist.

        This sets and overrides the default ``thumbnail`` and ``thumb`` properties.

        .. note::

            This method uses an API request to fetch the thumbnail.

        Returns
        -------
        str
            The URL to the video thumbnail.
        """
        if not node:
            node = NodePool.get_node()

        session: aiohttp.ClientSession = node._session
        url: str = f"https://img.youtube.com/vi/{self.identifier}/maxresdefault.jpg"

        async with session.get(url=url) as resp:
            if resp.status == 404:
                url = f'https://img.youtube.com/vi/{self.identifier}/hqdefault.jpg'

        self._thumb = url
        return url


class YouTubeMusicTrack(YouTubeTrack):
    """A track created using a search to YouTube Music."""

    PREFIX: str = "ytmsearch:"


class SoundCloudTrack(Playable):
    """A track created using a search to SoundCloud."""

    PREFIX: str = "scsearch:"


class YouTubePlaylist(Playable, Playlist):
    """Represents a Lavalink YouTube playlist object.

    Attributes
    ----------
    name: str
        The name of the playlist.
    tracks: :class:`YouTubeTrack`
        The list of :class:`YouTubeTrack` in the playlist.
    selected_track: Optional[int]
        The selected video in the playlist. This could be ``None``.
    """

    PREFIX: str = "ytpl:"

    def __init__(self, data: dict):
        self.tracks: list[YouTubeTrack] = []
        self.name: str = data["playlistInfo"]["name"]

        self.selected_track: Optional[int] = data["playlistInfo"].get("selectedTrack")
        if self.selected_track is not None:
            self.selected_track = int(self.selected_track)

        for track_data in data["tracks"]:
            track = YouTubeTrack(track_data)
            self.tracks.append(track)

    def __str__(self) -> str:
        return self.name



"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import Any

__all__ = (
    'WavelinkException',
    'AuthorizationFailed',
    'InvalidNode',
    'InvalidLavalinkVersion',
    'InvalidLavalinkResponse',
    'NoTracksError',
    'QueueEmpty',
    'InvalidChannelStateError',
    'InvalidChannelPermissions',
)


class WavelinkException(Exception):
    """Base wavelink exception."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args)


class AuthorizationFailed(WavelinkException):
    """Exception raised when password authorization failed for this Lavalink node."""
    pass


class InvalidNode(WavelinkException):
    pass


class InvalidLavalinkVersion(WavelinkException):
    """Exception raised when you try to use wavelink 2 with a Lavalink version under 3.7."""
    pass


class InvalidLavalinkResponse(WavelinkException):
    """Exception raised when wavelink receives an invalid response from Lavalink.

    Attributes
    ----------
    status: int | None
        The status code. Could be None.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args)
        self.status: int | None = kwargs.get('status')


class NoTracksError(WavelinkException):
    """Exception raised when no tracks could be found."""
    pass


class QueueEmpty(WavelinkException):
    """Exception raised when you try to retrieve from an empty queue."""
    pass


class InvalidChannelStateError(WavelinkException):
    """Base exception raised when an error occurs trying to connect to a :class:`discord.VoiceChannel`."""


class InvalidChannelPermissions(InvalidChannelStateError):
    """Exception raised when the client does not have correct permissions to join the channel.

    Could also be raised when there are too many users already in a user limited channel.
    """